unit SelDirDlg;

interface

uses
  Windows, SysUtils, Classes, FileCtrl;

type
  TDirName = type string;
  TAcceptDirEvent = procedure(Sender : TObject; Directory : TDirName) of object;

  TSelectDirDialog = class(TComponent)
  private
    FCaption : string;
    FRoot : TDirName;
    FDirectory : TDirName;
    FOptions : TSelectDirExtOpts;

    FOnAccept : TAcceptDirEvent;
    FOnCancel : TNotifyEvent;

    // Ces trois mthodes servent  indiquer si les proprits Caption, Root et Dir
    // doivent tre enregistre dans le fichier .dfm
    function CaptionStored : boolean;
    function RootStored : boolean;
    function DirStored : boolean;
  public
    constructor Create(AOwner : TComponent); override;

    // Affiche la bote de dialogue ; renvoie True si l'utilisateur clique sur OK
    function Execute : boolean;
  published
    // Titre de la bote de dialogue
    property Caption : string read FCaption write FCaption stored CaptionStored;
    // Rpertoire racine de l'arbre dans la bote de dialogue
    property Root : TDirName read FRoot write FRoot stored RootStored;
    // Rpertoire slectionn (avant et aprs) dans la bote de dialogue
    property Directory : TDirName read FDirectory write FDirectory stored DirStored;
    // Diverses options pour la bote de dialogue
    property Options : TSelectDirExtOpts read FOptions write FOptions default [sdNewUI];

    // Dclench lorsque l'utilisateur clique sur OK dans la bote de dialogue
    property OnAccept : TAcceptDirEvent read FOnAccept write FOnAccept;
    // Dclench lorsque l'utilisateur clique sur Annuler dans la bote de dialogue
    property OnCancel : TNotifyEvent read FOnCancel write FOnCancel;
  end;

implementation

var
  MainDisk : TDirName;

constructor TSelectDirDialog.Create(AOwner : TComponent);
begin
  inherited;
  FCaption := 'Slectionnez un dossier';
  FRoot := MainDisk;
  FDirectory := MainDisk;
  FOptions := [sdNewUI];
  FOnAccept := nil;
  FOnCancel := nil;
end;

function TSelectDirDialog.CaptionStored : boolean;
begin
  Result := FCaption <> 'Slectionnez un dossier';
end;

function TSelectDirDialog.RootStored : boolean;
begin
  Result := FRoot <> MainDisk;
end;

function TSelectDirDialog.DirStored : boolean;
begin
  Result := FDirectory <> MainDisk;
end;

function TSelectDirDialog.Execute : boolean;
begin
  Result := SelectDirectory(FCaption, FRoot, string(FDirectory), FOptions);
  if Result then
  begin
    if Assigned(FOnAccept) then
      FOnAccept(Self, FDirectory);
  end else
  begin
    if Assigned(FOnCancel) then
      FOnCancel(Self);
  end;
end;

initialization
  SetLength(MainDisk, MAX_PATH);
  SetLength(MainDisk, GetWindowsDirectory(PChar(MainDisk), MAX_PATH));
  MainDisk := IncludeTrailingPathDelimiter(ExtractFileDrive(MainDisk));
end.
